/*
 * parser.c:
 *      the parser used to interprete the high
 *      level data declarations; now builds the AST
 *
 ******* Grammar for Declarations
 *
 *  declns = decln* tEOF
 *  decln  = tID tEQ shapes tSEMI
 *  shapes = shape (tOR shape)*
 *  shape  = tID [ tOPENBR params tCLOSEBR ]
 *  params = param (tCOMMA param)*
 *  param  = tID tID
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "bool.h"
#include "lexer.h"
#include "ast.h"
#include "parser.h"

static void error( char *s )
{
	fprintf( stderr, "%s at line %d\n", s, lineno );
}

#define COPYOF(new,old) {new=malloc(1+strlen(old));if(new)strcpy(new,old);}
#define MUSTBE(t,mesg)	if( nexttok() != (t) ) {error(mesg); return FALSE;}

/* declns = decln* tEOF */
BOOL parse_declns( declnlist *d )
{
	*d = NULL;
	while( nexttok() == tID )    /* NB: a decln starts with a tID */
	{
		ungettok();
		if( ! parse_decln( d ) ) return FALSE;
		d = &( (*d)->next );
	}
	ungettok();
	MUSTBE( tEOF, "Spurious characters found beyond EOF" );
	return TRUE;
}

/* decln  = tID tEQ shapes tSEMI */
BOOL parse_decln( declnlist *d )
{
	char      *name;
	shapelist s;
	MUSTBE( tID, "declaration expected" );
	COPYOF( name, lexidval );
	MUSTBE( tEQ, "'=' expected" );
	if( ! parse_shapes(&s) ) return FALSE;
	MUSTBE( tSEMI, "'or' or ';' expected" );
	*d = build_declnlist( name, s, NULL );
	return TRUE;
}

/* shapes = shape (tOR shape)* */
BOOL parse_shapes( shapelist *s )
{
	for(;;) {
		if( ! parse_shape( s ) ) return FALSE;
	if( nexttok() != tOR ) break;
		s = &( (*s)->next );
	}
	ungettok();
	return TRUE;
}

/* shape = tID [ tOPENBR params tCLOSEBR ] */
BOOL parse_shape( shapelist *s )
{
	char *name;
	paramlist p = NULL;
	MUSTBE( tID, "shape name expected" );
	COPYOF( name, lexidval );
	if( nexttok() == tOPENBR )
	{
		if( ! parse_params(&p) ) return FALSE;
		MUSTBE( tCLOSEBR, "',' or ')' expected" );
	} else
	{
		ungettok();
	}
	*s = build_shapelist( name, p, NULL );
	return TRUE;
}

/* params = param (tCOMMA param)* */
/* param  = tID tID */
BOOL parse_params( paramlist * p )
{
	for(;;) {
		char *name;
		char *type;
		MUSTBE( tID, "Field type expected" );
		COPYOF( type, lexidval );
		MUSTBE( tID, "Field name expected" );
		COPYOF( name, lexidval );
		*p = build_paramlist( type, name, NULL );
	if( nexttok() != tCOMMA ) break;
		p = &( (*p)->next );
	}
	ungettok();
	return TRUE;
}
